use anyhow::Result;
use sqlx::{Connection, FromRow, Sqlite, Transaction};

use crate::repo::Storage;

#[derive(FromRow)]
pub struct StoredUser {
    pub id: u32,
    pub name: String,
    pub password: Option<String>,
    pub argon2_hash: Option<Box<str>>,
}

impl Storage {
    #[tracing::instrument(skip(self), name = "Storage::retrieve_user_by_name")]
    pub async fn retrieve_user_by_name(&self, name: &str) -> Result<Option<StoredUser>> {
        let mut executor = self.conn.lock().await;
        let res = sqlx::query_as(
            "select u.id, u.name, c.password, a.hash as argon2_hash
            from users u left join challenges_plain_password c on u.id = c.user_id
            left join challenges_argon2_password a on u.id = a.user_id
            where u.name = ?;",
        )
        .bind(name)
        .fetch_optional(&mut *executor)
        .await?;

        Ok(res)
    }

    #[tracing::instrument(skip(self), name = "Storage::check_user_existence")]
    pub async fn check_user_existence(&self, username: &str) -> Result<bool> {
        let mut executor = self.conn.lock().await;
        let result: Option<(String,)> = sqlx::query_as("select name from users where name = ?;")
            .bind(username)
            .fetch_optional(&mut *executor)
            .await?;

        Ok(result.is_some())
    }

    #[tracing::instrument(skip(self), name = "Storage::create_user")]
    pub async fn create_user(&self, name: &str) -> Result<()> {
        let query = sqlx::query(
            "insert into users(name)
            values (?);",
        )
        .bind(name);
        let mut executor = self.conn.lock().await;
        query.execute(&mut *executor).await?;

        Ok(())
    }

    #[tracing::instrument(skip(self, pwd), name = "Storage::set_password")]
    pub async fn set_password(&self, name: &str, pwd: &str) -> Result<Option<()>> {
        async fn inner(txn: &mut Transaction<'_, Sqlite>, name: &str, pwd: &str) -> Result<Option<()>> {
            let id: Option<(u32,)> = sqlx::query_as("select * from users where name = ? limit 1;")
                .bind(name)
                .fetch_optional(&mut **txn)
                .await?;
            let Some((id,)) = id else {
                return Ok(None);
            };
            sqlx::query("insert or replace into challenges_plain_password(user_id, password) values (?, ?);")
                .bind(id)
                .bind(pwd)
                .execute(&mut **txn)
                .await?;
            Ok(Some(()))
        }

        let mut executor = self.conn.lock().await;
        let mut tx = executor.begin().await?;
        let res = inner(&mut tx, name, pwd).await;
        match res {
            Ok(e) => {
                tx.commit().await?;
                Ok(e)
            }
            Err(e) => {
                tx.rollback().await?;
                Err(e)
            }
        }
    }

    #[tracing::instrument(skip(self), name = "Storage::retrieve_user_id_by_name")]
    pub async fn retrieve_user_id_by_name(&self, name: &str) -> Result<Option<u32>> {
        let mut executor = self.conn.lock().await;
        let res: Option<(u32,)> = sqlx::query_as("select u.id from users u where u.name = ?;")
            .bind(name)
            .fetch_optional(&mut *executor)
            .await?;

        Ok(res.map(|(id,)| id))
    }

    #[tracing::instrument(skip(self), name = "Storage::create_or_retrieve_user_id_by_name")]
    pub async fn create_or_retrieve_user_id_by_name(&self, name: &str) -> Result<u32> {
        let mut executor = self.conn.lock().await;
        let res: (u32,) = sqlx::query_as(
            "insert into users(name)
            values (?)
            on conflict(name) do update set name = excluded.name
            returning id;",
        )
        .bind(name)
        .fetch_one(&mut *executor)
        .await?;

        Ok(res.0)
    }
}
